#include "camera/camera.h"
#include "list/typed_list.h"
#include "mem_arena.h"
#include "mem_utils.h"
#include "misc/misc_utils.h"
#include "rasteriser/rasteriser.h"
#include "scene/scene.h"
#include "vector/vec.h"
#include "window/window.h"
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_keycode.h>
#include <stdlib.h>

int main(void) {
  colour_t bg =
      (colour_t){.rgba.r = 27, .rgba.g = 38, .rgba.b = 79, .rgba.a = 255};
  camera_t camera = {
      .position = {.x = 0.0f, .y = 0.0f, .z = 0.0f},
      .rotation = {.x = 0.0f, .y = 0.0f, .z = 0.0f},
  };
  vec3f_t viewport = {.x = 1.0f, .y = 1.0f, .z = 1.0f};

  window_t window = {0};

  if (!init_window(&window, 800, 800, "CG From Scratch Rasteriser")) {
    return EXIT_FAILURE;
  }

  Arena *arena = NULL;
  u64 main_arena_capacity = 20ull * 1024ull * 1024ull * 1024ull;
  if (!wapp_mem_arena_init(&arena, main_arena_capacity, WAPP_MEM_ALLOC_RESERVE,
                           false)) {
    return EXIT_FAILURE;
  }

  vec3f_t translation = {-1.5f, 0.0f, 7.0f};

  list_vec3f_t *vertices = list_create(vec3f_t, arena);
  if (!vertices) {
    return EXIT_FAILURE;
  }
  // Vertices
  vec3f_t v0 = vec_add(vec3f_t, translation, ((vec3f_t){1, 1, 1}));
  vec3f_t v1 = vec_add(vec3f_t, translation, ((vec3f_t){-1, 1, 1}));
  vec3f_t v2 = vec_add(vec3f_t, translation, ((vec3f_t){-1, -1, 1}));
  vec3f_t v3 = vec_add(vec3f_t, translation, ((vec3f_t){1, -1, 1}));
  vec3f_t v4 = vec_add(vec3f_t, translation, ((vec3f_t){1, 1, -1}));
  vec3f_t v5 = vec_add(vec3f_t, translation, ((vec3f_t){-1, 1, -1}));
  vec3f_t v6 = vec_add(vec3f_t, translation, ((vec3f_t){-1, -1, -1}));
  vec3f_t v7 = vec_add(vec3f_t, translation, ((vec3f_t){1, -1, -1}));

  list_append(vec3f_t, arena, vertices, v0);
  list_append(vec3f_t, arena, vertices, v1);
  list_append(vec3f_t, arena, vertices, v2);
  list_append(vec3f_t, arena, vertices, v3);
  list_append(vec3f_t, arena, vertices, v4);
  list_append(vec3f_t, arena, vertices, v5);
  list_append(vec3f_t, arena, vertices, v6);
  list_append(vec3f_t, arena, vertices, v7);

  list_scene_triangle_t *triangles = list_create(scene_triangle_t, arena);
  if (!triangles) {
    return EXIT_FAILURE;
  }
  // Triangles
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){0, 1, 2, (colour_t){.colour = 0xff0000ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){0, 2, 3, (colour_t){.colour = 0xff0000ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){4, 0, 3, (colour_t){.colour = 0x00ff00ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){4, 3, 7, (colour_t){.colour = 0x00ff00ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){5, 4, 7, (colour_t){.colour = 0x0000ffff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){5, 7, 6, (colour_t){.colour = 0x0000ffff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){1, 5, 6, (colour_t){.colour = 0xffff00ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){1, 6, 2, (colour_t){.colour = 0xffff00ff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){4, 5, 1, (colour_t){.colour = 0xff00ffff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){4, 1, 0, (colour_t){.colour = 0xff00ffff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){2, 6, 7, (colour_t){.colour = 0x00ffffff}}));
  list_append(scene_triangle_t, arena, triangles,
              ((scene_triangle_t){2, 7, 3, (colour_t){.colour = 0x00ffffff}}));

  bool running = true;
  SDL_Event event = {0};

  // i32 w_min = ((i32)window.half_width) * -1;
  // i32 w_max = (i32)window.half_width;
  // i32 h_min = ((i32)window.half_height) * -1;
  // i32 h_max = (i32)window.half_height;

  Arena *frame_arena = NULL;
  u64 frame_arena_capacity = 10ull * 1024ull * 1024ull * 1024ull;
  if (!wapp_mem_arena_init(&frame_arena, frame_arena_capacity,
                           WAPP_MEM_ALLOC_RESERVE, false)) {
    return EXIT_FAILURE;
  }

  while (running) {
    while (SDL_PollEvent(&event)) {
      switch (event.type) {
      case SDL_QUIT:
        running = false;
        break;
      }
    }

    clear_window(&window, bg);

    render_object(&window, arena, &camera, viewport, vertices, triangles);

    swap_buffers(&window);

    wapp_mem_arena_clear(frame_arena);
  }

  wapp_mem_arena_destroy(&frame_arena);
  wapp_mem_arena_destroy(&arena);
  close_window(&window);

  return EXIT_SUCCESS;
}
